# pylint: disable=no-self-argument

import sys
from io import BufferedIOBase

# Disabling pylint no-name-in-module because this is the documented way to
# import `BaseModel` and it's not broken, so let's keep it.
from pydantic import (  # pylint: disable=no-name-in-module
    Field,
    FilePath,
    HttpUrl,
    root_validator,
    validator,
)
from typing import Any, Callable, Dict, List, Optional, Text, Tuple, Union

from .base import BaseModel


if sys.version_info >= (3, 8):
    from typing import Literal  # pylint: disable=no-name-in-module
else:
    from typing_extensions import Literal


class EmbeddedLogo(BaseModel):
    src: str
    max_width: Optional[int]
    max_height: Optional[int]


class FillPDFPayload(BaseModel):
    data: Union[List[Dict[str, Any]], Dict[str, Any]]
    title: Optional[str] = None
    font_size: Optional[int] = None
    text_color: Optional[str] = None

    @validator("data")
    def data_cannot_be_empty(cls, v):
        if isinstance(v, dict) and len(v) == 0:
            raise ValueError("cannot be empty")
        return v


class GeneratePDFPayload(BaseModel):
    data: Union[List[Dict[str, Any]], Dict[Literal["html", "css"], str]]
    logo: Optional[EmbeddedLogo]
    title: Optional[str]
    type: Optional[Literal["markdown", "html"]] = "markdown"

    @validator("data")
    def data_must_match_type(cls, v, values):
        if "type" in values and values["type"] == "html":
            if "html" not in v:
                raise ValueError("must contain HTML if using `html` type")
            if isinstance(v, list):
                raise ValueError("must contain a data dict not a list")
        return v

    @validator("type")
    def type_must_match_data(cls, v, values):
        if v == "html":
            if "data" in values and not isinstance(values["data"], dict):
                raise ValueError("HTML types must used a dict data payload")
        return v


class GenerateEtchSigningURLPayload(BaseModel):
    signer_eid: str
    client_user_id: str


class SignerField(BaseModel):
    file_id: str
    field_id: str


class EtchSigner(BaseModel):
    """Dataclass representing etch signers."""

    name: str
    email: str
    fields: List[SignerField]
    signer_type: str = "email"
    # id will be generated if `None`
    id: Optional[str] = None
    routing_order: Optional[int] = None
    redirect_url: Optional[str] = Field(None, alias="redirectURL")
    accept_each_field: Optional[bool] = None
    enable_emails: Optional[List[str]] = None
    # signature_mode can be "draw" or "text" (default: text)
    signature_mode: Optional[str] = None


class SignatureField(BaseModel):
    id: str
    type: str
    page_num: int
    # Should be in a format similar to:
    # { x: 100.00, y: 121.21, width: 33.00 }
    rect: Dict[str, float]


class Base64Upload(BaseModel):
    data: str
    filename: str
    mimetype: str = "application/pdf"


class TableColumnAlignment(BaseModel):
    align: Optional[Literal["left", "center", "right"]] = None
    width: Optional[str] = None


# https://www.useanvil.com/docs/api/generate-pdf#table
class MarkdownTable(BaseModel):
    rows: List[List[str]]

    # defaults to `True` if not provided.
    # set to false for no header row on the table
    first_row_headers: Optional[bool] = None

    # defaults to `False` if not provided.
    # set to true to display gridlines in-between rows or columns
    row_grid_lines: Optional[bool] = True
    column_grid_lines: Optional[bool] = False

    # defaults to 'top' if not provided.
    # adjust vertical alignment of table text
    # accepts 'top', 'center', or 'bottom'
    vertical_align: Optional[Literal["top", "center", "bottom"]] = "center"

    # (optional) columnOptions - An array of columnOption objects.
    # You do not need to specify all columns. Accepts an
    # empty object indicating no overrides on the
    # specified column.
    #
    # Supported keys for columnOption:
    # align (optional) - adjust horizontal alignment of table text
    # accepts 'left', 'center', or 'right'; defaults to 'left'
    # width (optional) - adjust the width of the column
    # accepts width in pixels or as percentage of the table width
    column_options: Optional[List[TableColumnAlignment]] = None


# https://www.useanvil.com/docs/api/object-references/#verbatimfield
class MarkdownContent(BaseModel):
    label: Optional[str] = None
    heading: Optional[str] = None
    content: Optional[str] = None
    table: Optional[MarkdownTable] = None
    font_size: int = 14
    text_color: str = "#000000"


class DocumentMarkup(BaseModel):
    """Dataclass representing a document with HTML/CSS markup."""

    id: str
    filename: str
    markup: Dict[Literal["html", "css"], str]
    fields: Optional[List[SignatureField]] = None
    title: Optional[str] = None
    font_size: int = 14
    text_color: str = "#000000"


class DocumentMarkdown(BaseModel):
    """Dataclass representing a document with Markdown."""

    id: str
    filename: str
    # NOTE: Order matters here in the Union[].
    # If `SignatureField` is not first, the types are similar enough that it
    # will use `MarkdownContent` instead.
    fields: Optional[List[Union[SignatureField, MarkdownContent]]] = None
    title: Optional[str] = None
    font_size: int = 14
    text_color: str = "#000000"


class DocumentUpload(BaseModel):
    """Dataclass representing an uploaded document."""

    id: str
    title: str
    file: "UploadableFile"
    fields: List[SignatureField]
    font_size: int = 14
    text_color: str = "#000000"

    class Config:
        """Special rules for this model class."""

        arbitrary_types_allowed = True
        # String to help point out the file upload attr.
        # Used in multipart uploads and should not affect base64-encoded files.
        contains_uploads = "file"


class EtchCastRef(BaseModel):
    """Dataclass representing an existing template used as a reference."""

    id: str
    cast_eid: str


class CreateEtchFilePayload(BaseModel):
    payloads: Union[str, Dict[str, FillPDFPayload]]


class CreateEtchPacketPayload(BaseModel):
    """
    Payload for createEtchPacket.

    See the full packet payload defined here:
    https://www.useanvil.com/docs/api/e-signatures#tying-it-all-together
    """

    name: str
    signers: List[EtchSigner]
    files: List["AttachableEtchFile"]
    signature_email_subject: Optional[str] = None
    signature_email_body: Optional[str] = None
    is_draft: Optional[bool] = False
    is_test: Optional[bool] = True
    merge_pdfs: Optional[bool] = Field(None, alias="mergePDFs")
    data: Optional[CreateEtchFilePayload] = None
    signature_page_options: Optional[Dict[Any, Any]] = None
    webhook_url: Optional[str] = Field(None, alias="webhookURL")
    reply_to_name: Optional[Any] = None
    reply_to_email: Optional[Any] = None
    enable_emails: Optional[Union[bool, List[str]]] = None
    create_cast_templates_from_uploads: Optional[bool] = None
    duplicate_casts: Optional[bool] = None


class ForgeSubmitPayload(BaseModel):
    """
    Payload for forgeSubmit.

    See full payload defined here:
    https://www.useanvil.com/docs/api/graphql/reference/#operation-forgesubmit-Mutations
    """

    forge_eid: Text
    payload: Dict[Text, Any]
    weld_data_eid: Optional[Text] = None
    submission_eid: Optional[Text] = None
    # Defaults to True when not provided/is None
    enforce_payload_valid_on_create: Optional[bool] = None
    current_step: Optional[int] = None
    complete: Optional[bool] = None
    # Note that if using a development API key, this will be forced to `True`
    # even when `False` is used in the payload.
    is_test: Optional[bool] = True
    timezone: Optional[Text] = None
    webhook_url: Optional[HttpUrl] = Field(None, alias="webhookURL")
    group_array_id: Optional[Text] = None
    group_array_index: Optional[int] = None

    @root_validator
    def wd_submission_both_required(cls, values):
        both_required = ["weld_data_eid", "submission_eid"]
        picked = [k for k in both_required if k in values and values[k] is not None]
        if len(picked) == 1:
            raise ValueError(
                "Both `weld_data_eid` and `submission_eid` are "
                "required if either are provided."
            )
        return values


# Types usable for uploads:
# Base64Upload: An entire PDF file base64 encoded
# FilePath: The absolute path of the file. When the pydantic model is used,
#   it will convert a path string (i.e. "/home/me/my_file.pdf") into a
#   `pathlib.Path` instance.
# UploadMethod: A function that returns a tuple of:
#   * BufferedIOBase instance for use in `requests`
#   * String containing the name for the file for use in Anvil
UploadMethod = Callable[[Any], Tuple[BufferedIOBase, str]]
UploadableFile = Union[Base64Upload, FilePath, UploadMethod]
AttachableEtchFile = Union[
    DocumentUpload, EtchCastRef, DocumentMarkup, DocumentMarkdown
]

# Classes below use types wrapped in quotes avoid a circular dependency/weird
# variable assignment locations with the aliases above. We need to manually
# update the refs for them to point to the right things.
DocumentUpload.update_forward_refs()
CreateEtchPacketPayload.update_forward_refs()



from typing import Any, Dict

from python_anvil.http import HTTPClient


class AnvilRequest:
    show_headers = False
    _client: HTTPClient

    def get_url(self):
        raise NotImplementedError

    def _request(self, method, url, **kwargs):
        if not self._client:
            raise AssertionError(
                "Client has not been initialized. Please use the constructors "
                "provided by the other request implementations."
            )

        if method.upper() not in ["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"]:
            raise ValueError("Invalid HTTP method provided")

        full_url = "/".join([self.get_url(), url]) if len(url) > 0 else self.get_url()

        return self._client.request(method, full_url, **kwargs)

    def handle_error(self, response, status_code, headers):
        extra = None
        if self.show_headers:
            extra = headers

        if hasattr(response, "decode"):
            message = f"Error: {status_code}: {response.decode()} {extra}"
        else:
            message = f"Error: {status_code}: {response} {extra}"

        raise Exception(message)

    def process_response(self, response, status_code, headers, **kwargs):
        res = response
        if not 200 <= status_code < 300:
            self.handle_error(response, status_code, headers)

        debug = kwargs.pop("debug", False)

        # Include headers alongside the response.
        # This is useful for figuring out rate limits outside of this library's
        # scope and to manage waiting.
        include_headers = kwargs.pop("include_headers", False)
        if debug or include_headers:
            return {"response": res, "headers": headers}

        return res


class BaseAnvilHttpRequest(AnvilRequest):
    def __init__(self, client, options=None):
        self._client = client
        self._options = options

    def get_url(self):
        raise NotImplementedError

    def get(self, url, params=None, **kwargs):
        retry = kwargs.pop("retry", True)
        content, status_code, headers = self._request(
            "GET", url, params=params, retry=retry
        )
        return self.process_response(content, status_code, headers, **kwargs)

    def post(self, url, data=None, **kwargs):
        retry = kwargs.pop("retry", True)
        params = kwargs.pop("params", None)
        content, status_code, headers = self._request(
            "POST", url, json=data, retry=retry, params=params
        )
        return self.process_response(content, status_code, headers, **kwargs)


class GraphqlRequest(AnvilRequest):
    API_HOST = "https://graphql.useanvil.com"

    def __init__(self, client: HTTPClient):
        self._client = client

    def get_url(self):
        return f"{self.API_HOST}"

    def post(self, query, variables=None, **kwargs):
        return self.run_query("POST", query, variables=variables, **kwargs)

    def post_multipart(self, files=None, **kwargs):
        return self.run_query("POST", None, files=files, is_multipart=True, **kwargs)

    def run_query(
        self, method, query, variables=None, files=None, is_multipart=False, **kwargs
    ):
        if not query and not files:
            raise AssertionError(
                "Either `query` or `files` must be passed into this method."
            )
        data: Dict[str, Any] = {}

        if query:
            data["query"] = query

        if files and is_multipart:
            # Make sure `data` is nothing when we're doing a multipart request.
            data = {}
        elif variables:
            data["variables"] = variables

        # Optional debug kwargs.
        # At this point, only the CLI will pass this in as a
        # "show me everything" sort of switch.
        debug = kwargs.pop("debug", False)
        include_headers = kwargs.pop("include_headers", False)

        content, status_code, headers = self._request(
            method,
            # URL blank here since graphql requests don't append to url
            '',
            # Queries need to be wrapped by curly braces(?) based on the
            # current API implementation.
            # The current library for graphql query generation doesn't do this(?)
            json=data,
            files=files,
            parse_json=True,
            **kwargs,
        )

        return self.process_response(
            content,
            status_code,
            headers,
            debug=debug,
            include_headers=include_headers,
            **kwargs,
        )


class RestRequest(BaseAnvilHttpRequest):
    API_HOST = "https://app.useanvil.com"
    API_BASE = "api"
    API_VERSION = "v1"

    def get_url(self):
        return f"{self.API_HOST}/{self.API_BASE}/{self.API_VERSION}"


class PlainRequest(BaseAnvilHttpRequest):
    API_HOST = "https://app.useanvil.com"
    API_BASE = "api"

    def get_url(self):
        return f"{self.API_HOST}/{self.API_BASE}"



import click
import os
from csv import DictReader
from logging import getLogger
from tabulate import tabulate
from time import sleep
from typing import List

from python_anvil import utils

from .api import Anvil
from .api_resources.payload import FillPDFPayload


logger = getLogger(__name__)


def get_api_key():
    return os.environ.get("ANVIL_API_KEY")


def contains_headers(res):
    return isinstance(res, dict) and "headers" in res


def process_response(res):
    return res["response"], res["headers"]


@click.group()
@click.option("--debug/--no-debug", default=False)
@click.pass_context
def cli(ctx: click.Context, debug=False):
    ctx.ensure_object(dict)

    key = get_api_key()
    if not key:
        raise ValueError("$ANVIL_API_KEY must be defined in your environment variables")

    anvil = Anvil(key)
    ctx.obj["anvil"] = anvil
    ctx.obj["debug"] = debug


@cli.command("current-user", help="Show details about your API user")
@click.pass_context
def current_user(ctx):
    anvil = ctx.obj["anvil"]
    debug = ctx.obj["debug"]
    res = anvil.get_current_user(debug=debug)

    if contains_headers(res):
        res, headers = process_response(res)
        if debug:
            click.echo(headers)

    click.echo(f"User data: \n\n {res}")


@cli.command()
@click.option(
    "-i", "-in", "input_filename", help="Filename of input payload", required=True
)
@click.option(
    "-o",
    "--out",
    "out_filename",
    help="Filename of output PDF",
    required=True,
)
@click.pass_context
def generate_pdf(ctx, input_filename, out_filename):
    """Generate a PDF."""
    anvil = ctx.obj["anvil"]
    debug = ctx.obj["debug"]

    with click.open_file(input_filename, "r") as infile:
        res = anvil.generate_pdf(infile.read(), debug=debug)

    if contains_headers(res):
        res, headers = process_response(res)
        if debug:
            click.echo(headers)

    with click.open_file(out_filename, "wb") as file:
        file.write(res)


@cli.command()
@click.option("-l", "--list", "list_all", help="List all available welds", is_flag=True)
@click.argument("eid", default="")
@click.pass_context
def weld(ctx, eid, list_all):
    """Fetch weld info or list of welds."""
    anvil = ctx.obj["anvil"]
    debug = ctx.obj["debug"]

    if list_all:
        res = anvil.get_welds(debug=debug)
        if contains_headers(res):
            res, headers = process_response(res)
            if debug:
                click.echo(headers)

        data = [(w["eid"], w.get("slug"), w.get("title"), w.get("forges")) for w in res]
        click.echo(tabulate(data, tablefmt="pretty", headers=["eid", "slug", "title"]))
        return

    if not eid:
        raise Exception("You need to pass in a weld eid")

    res = anvil.get_weld(eid)
    print(res)


@cli.command()
@click.option(
    "-l",
    "--list",
    "list_templates",
    help="List available casts marked as templates",
    is_flag=True,
)
@click.option(
    "-a", "--all", "list_all", help="List all casts, even non-templates", is_flag=True
)
@click.option("--version_number", help="Get the specified version of this cast")
@click.argument("eid", default="")
@click.pass_context
def cast(ctx, eid, version_number, list_all, list_templates):
    """Fetch Cast data given a Cast eid."""
    anvil = ctx.obj["anvil"]  # type: Anvil
    debug = ctx.obj["debug"]

    if not eid and not (list_templates or list_all):
        raise AssertionError("Cast eid or --list/--all option required")

    if list_all or list_templates:
        res = anvil.get_casts(show_all=list_all, debug=debug)

        if contains_headers(res):
            res, headers = process_response(res)
            if debug:
                click.echo(headers)

        data = [[c["eid"], c["title"]] for c in res]
        click.echo(tabulate(data, headers=["eid", "title"]))
        return

    if eid:
        click.echo(f"Getting cast with eid '{eid}' \n")
        _res = anvil.get_cast(eid, version_number=version_number, debug=debug)

        if contains_headers(_res):
            res, headers = process_response(res)
            if debug:
                click.echo(headers)

        def get_field_info(cc):
            return tabulate(cc.get("fields", []))

        if not _res:
            click.echo(f"Cast with eid: {eid} not found")
            return

        table_data = [[_res["eid"], _res["title"], get_field_info(_res["fieldInfo"])]]
        click.echo(tabulate(table_data, tablefmt="pretty", headers=list(_res.keys())))


@cli.command("fill-pdf")
@click.argument("template_id")
@click.option(
    "-o",
    "--out",
    "out_filename",
    required=True,
    help="Filename of output PDF",
)
@click.option(
    "-i",
    "--input",
    "payload_csv",
    required=True,
    help="Filename of input CSV that provides data",
)
@click.pass_context
def fill_pdf(ctx, template_id, out_filename, payload_csv):
    """Fill PDF template with data."""
    anvil = ctx.obj["anvil"]
    debug = ctx.obj["debug"]

    if all([template_id, out_filename, payload_csv]):
        payloads = []  # type: List[FillPDFPayload]
        with click.open_file(payload_csv, "r") as csv_file:
            reader = DictReader(csv_file)
            # NOTE: This is potentially problematic for larger datasets and/or
            # very long csv files, but not sure if the use-case is there yet..
            #
            # Once memory/execution times are a problem for this command, the
            # `progressbar()` can be removed below and we could just work on
            # each csv line individually without loading it all into memory
            # as we are doing here (or with `list()`). But then that removes
            # the nice progress bar, so..trade-offs!
            for row in reader:
                payloads.append(FillPDFPayload(data=dict(row)))

        with click.progressbar(payloads, label="Filling PDFs and saving") as ps:
            indexed_files = utils.build_batch_filenames(out_filename)
            for payload in ps:
                res = anvil.fill_pdf(template_id, payload.to_dict(), debug=debug)

                if contains_headers(res):
                    res, headers = process_response(res)
                    if debug:
                        click.echo(headers)

                next_file = next(indexed_files)
                click.echo(f"\nWriting {next_file}")
                with click.open_file(next_file, "wb") as file:
                    file.write(res)
                sleep(1)


@cli.command("create-etch")
@click.option(
    "-p",
    "--payload",
    "payload",
    type=click.File('rb'),
    required=True,
    help="File that contains JSON payload",
)
@click.pass_context
def create_etch(ctx, payload):
    """Create an etch packet with a JSON file.

    Example usage:
        # For existing files
        > $ ANVIL_API_KEY=mykey anvil create-etch --payload=my_payload_file.json

        # You can also get data from STDIN
        > $ ANVIL_API_KEY=mykey anvil create-etch --payload -
    """
    anvil = ctx.obj["anvil"]
    debug = ctx.obj["debug"]
    res = anvil.create_etch_packet(json=payload.read(), debug=debug)

    if contains_headers(res):
        res, headers = process_response(res)
        if debug:
            click.echo(headers)

    if "data" in res:
        click.echo(
            f"Etch packet created with id: {res['data']['createEtchPacket']['eid']}"
        )
    else:
        click.echo(res)


@cli.command("generate-etch-url", help="Generate an etch url for a signer")
@click.option(
    "-c",
    "--client",
    "client_user_id",
    required=True,
    help="The signer's user id in your system belongs here",
)
@click.option(
    "-s",
    "--signer",
    "signer_eid",
    required=True,
    help="The eid of the next signer belongs here. The signer's eid can be "
    "found in the response of the `createEtchPacket` mutation",
)
@click.pass_context
def generate_etch_url(ctx, signer_eid, client_user_id):
    anvil = ctx.obj["anvil"]
    debug = ctx.obj["debug"]
    res = anvil.generate_etch_signing_url(
        signer_eid=signer_eid, client_user_id=client_user_id, debug=debug
    )

    if contains_headers(res):
        res, headers = process_response(res)
        if debug:
            click.echo(headers)

    url = res.get("data", {}).get("generateEtchSignURL")
    click.echo(f"Signing URL is: {url}")


@cli.command("download-documents", help="Download etch documents")
@click.option(
    "-d",
    "--document-group",
    "document_group_eid",
    required=True,
    help="The documentGroupEid can be found in the response of the "
    "createEtchPacket or sendEtchPacket mutations.",
)
@click.option(
    "-f", "--filename", "filename", help="Optional filename for the downloaded zip file"
)
@click.option(
    "--stdout/--no-stdout",
    help="Instead of writing to a file, output data to STDOUT",
    default=False,
)
@click.pass_context
def download_documents(ctx, document_group_eid, filename, stdout):
    anvil = ctx.obj["anvil"]
    debug = ctx.obj["debug"]
    res = anvil.download_documents(document_group_eid, debug=debug)

    if contains_headers(res):
        res, headers = process_response(res)
        if debug:
            click.echo(headers)

    if not stdout:
        if not filename:
            filename = f"{document_group_eid}.zip"

        with click.open_file(filename, 'wb') as out_file:
            out_file.write(res)
        click.echo(f"Saved as '{click.format_filename(filename)}'")
    else:
        click.echo(res)


@cli.command('gql-query', help="Run a raw graphql query")
@click.option(
    "-q",
    "--query",
    "query",
    required=True,
    help="The query body. This is the 'query' part of the JSON payload",
)
@click.option(
    "-v",
    "--variables",
    "variables",
    help="The query variables. This is the 'variables' part of the JSON payload",
)
@click.pass_context
def gql_query(ctx, query, variables):
    anvil = ctx.obj["anvil"]
    debug = ctx.obj["debug"]
    res = anvil.query(query, variables=variables, debug=debug)

    if contains_headers(res):
        res, headers = process_response(res)
        if debug:
            click.echo(headers)

    click.echo(res)


if __name__ == "__main__":  # pragma: no cover
    cli()  # pylint: disable=no-value-for-parameter



# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned

import json
import pytest
from click.testing import CliRunner
from unittest import mock

from python_anvil.cli import cli


@pytest.fixture
def runner():
    return CliRunner()


def set_key(monkeypatch):
    monkeypatch.setenv("ANVIL_API_KEY", "MY_KEY")


def describe_cli():
    @mock.patch("python_anvil.api.Anvil.get_current_user")
    def it_handles_no_key(anvil, runner):
        res = runner.invoke(cli, ["current-user"])
        assert anvil.call_count == 0
        assert isinstance(res.exception, ValueError)

    @mock.patch("python_anvil.api.Anvil.get_current_user")
    def it_handles_key(anvil, runner, monkeypatch):
        set_key(monkeypatch)
        res = runner.invoke(cli, ["current-user"])
        assert anvil.call_count == 1
        assert not isinstance(res.exception, ValueError)

    def describe_current_user():
        @mock.patch("python_anvil.api.Anvil.query")
        def it_queries(query, runner, monkeypatch):
            set_key(monkeypatch)
            query.return_value = {"data": {"currentUser": {"name": "Cameron"}}}

            res = runner.invoke(cli, ['current-user'])
            assert "{'name': 'Cameron'}" in res.output
            assert "User data:" in res.output
            assert query.call_count == 1

        @mock.patch("python_anvil.api.Anvil.query")
        def it_handles_headers(query, runner, monkeypatch):
            set_key(monkeypatch)
            query.return_value = {
                "response": {"data": {"currentUser": {"name": "Cameron"}}},
                "headers": {"Header-1": "val1", "Header-2": "val2"},
            }

            res = runner.invoke(cli, ['--debug', 'current-user'])
            assert "{'name': 'Cameron'}" in res.output
            assert "User data:" in res.output
            assert "{'Header-1': 'val1'," in res.output
            assert query.call_count == 1

    def describe_generate_pdf():
        @mock.patch("python_anvil.api.Anvil.generate_pdf")
        def it_handles_files(generate_pdf, runner, monkeypatch):
            set_key(monkeypatch)

            in_data = json.dumps({"data": "", "title": "My Title"})
            generate_pdf.return_value = "Some bytes"
            mock_open = mock.mock_open(read_data=in_data)

            with mock.patch("click.open_file", mock_open) as m:
                res = runner.invoke(
                    cli, ['generate-pdf', '-i', 'infile', '-o', 'outfile']
                )
                generate_pdf.assert_called_once_with(in_data, debug=False)
                m().write.assert_called_once_with("Some bytes")



# pylint: disable=too-many-instance-attributes
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from python_anvil.api_resources.mutations.base import BaseQuery
from python_anvil.api_resources.payload import (
    AttachableEtchFile,
    Base64Upload,
    CreateEtchFilePayload,
    CreateEtchPacketPayload,
    DocumentUpload,
    EtchSigner,
)
from python_anvil.utils import create_unique_id


DEFAULT_RESPONSE_QUERY = """{
  id
  eid
  name
  detailsURL
  documentGroup {
    id
    eid
    status
    files
    signers {
      id
      eid
      aliasId
      routingOrder
      name
      email
      status
      signActionType
    }
  }
}"""

# NOTE: Since the below will be used as a formatted string (this also applies
#   to f-strings) any literal curly braces need to be doubled, else they'll be
#   interpreted as string replacement tokens.
CREATE_ETCH_PACKET = """
mutation CreateEtchPacket (
    $name: String,
    $files: [EtchFile!],
    $isDraft: Boolean,
    $isTest: Boolean,
    $mergePDFs: Boolean,
    $signatureEmailSubject: String,
    $signatureEmailBody: String,
    $signatureProvider: String,
    $signaturePageOptions: JSON,
    $signers: [JSON!],
    $webhookURL: String,
    $replyToName: String,
    $replyToEmail: String,
    $data: JSON,
    $enableEmails: JSON,
    $createCastTemplatesFromUploads: Boolean,
    $duplicateCasts: Boolean=false,
  ) {{
    createEtchPacket (
      name: $name,
      files: $files,
      isDraft: $isDraft,
      isTest: $isTest,
      mergePDFs: $mergePDFs,
      signatureEmailSubject: $signatureEmailSubject,
      signatureEmailBody: $signatureEmailBody,
      signatureProvider: $signatureProvider,
      signaturePageOptions: $signaturePageOptions,
      signers: $signers,
      webhookURL: $webhookURL,
      replyToName: $replyToName,
      replyToEmail: $replyToEmail,
      data: $data,
      enableEmails: $enableEmails,
      createCastTemplatesFromUploads: $createCastTemplatesFromUploads,
      duplicateCasts: $duplicateCasts
    )
        {query}
  }}
"""


class CreateEtchPacket(BaseQuery):
    mutation = CREATE_ETCH_PACKET
    mutation_res_query = DEFAULT_RESPONSE_QUERY

    def __init__(  # pylint: disable=too-many-locals
        self,
        name: Optional[str] = None,
        signature_email_subject: Optional[str] = None,
        signature_email_body: Optional[str] = None,
        signers: Optional[List[EtchSigner]] = None,
        files: Optional[List[AttachableEtchFile]] = None,
        file_payloads: Optional[dict] = None,
        signature_page_options: Optional[Dict[Any, Any]] = None,
        is_draft: bool = False,
        is_test: bool = True,
        payload: Optional[CreateEtchPacketPayload] = None,
        webhook_url: Optional[str] = None,
        reply_to_name: Optional[str] = None,
        reply_to_email: Optional[str] = None,
        merge_pdfs: Optional[bool] = None,
        enable_emails: Optional[Union[bool, List[str]]] = None,
        create_cast_templates_from_uploads: Optional[bool] = None,
        duplicate_casts: Optional[bool] = None,
    ):
        # `name` is required when `payload` is not present.
        if not payload and not name:
            raise TypeError(
                "Missing 2 required positional arguments: 'name' and "
                "'signature_email_subject'"
            )

        self.name = name
        self.signature_email_subject = signature_email_subject
        self.signature_email_body = signature_email_body
        self.signature_page_options = signature_page_options
        self.signers = signers or []
        self.files = files or []
        self.file_payloads = file_payloads or {}
        self.is_draft = is_draft
        self.is_test = is_test
        self.payload = payload
        self.webhook_url = webhook_url
        self.reply_to_name = reply_to_name
        self.reply_to_email = reply_to_email
        self.merge_pdfs = merge_pdfs
        self.enable_emails = enable_emails
        self.create_cast_templates_from_uploads = create_cast_templates_from_uploads
        self.duplicate_casts = duplicate_casts

    @classmethod
    def create_from_dict(cls, payload: Dict) -> 'CreateEtchPacket':
        """Create a new instance of `CreateEtchPacket` from a dict payload."""
        try:
            mutation = cls(
                **{k: v for k, v in payload.items() if k not in ["signers", "files"]}
            )
        except TypeError as e:
            raise ValueError(
                f"`payload` must be a valid CreateEtchPacket instance or dict. {e}"
            ) from e
        if "signers" in payload:
            for signer in payload["signers"]:
                mutation.add_signer(EtchSigner(**signer))

        if "files" in payload:
            for file in payload["files"]:
                mutation.add_file(DocumentUpload(**file))

        return mutation

    def add_signer(self, signer: Union[dict, EtchSigner]):
        """Add a signer to the mutation payload.

        :param signer: Signer object to add to the payload
        :type signer: dict|EtchSigner
        """
        if isinstance(signer, dict):
            data = EtchSigner(**signer)
        elif isinstance(signer, EtchSigner):
            data = signer
        else:
            raise ValueError("Signer must be either a dict or EtchSigner type")

        if data.signer_type not in ["embedded", "email"]:
            raise ValueError(
                "Etch signer `signer_type` must be only 'embedded' or 'email"
            )

        if not data.id:
            data.id = create_unique_id("signer")
        if not data.routing_order:
            if self.signers:
                # Basic thing to get the next number
                # But this might not be necessary since API goes by index
                # of signers in the list.
                all_signers = [(s.routing_order or 0) for s in self.signers]
                num = max(all_signers) + 1
            else:
                num = 1
            data.routing_order = num

        self.signers.append(data)

    def add_file(self, file: AttachableEtchFile):
        """Add file to a pending list of Upload objects.

        Files will not be uploaded when running this method. They will be
        uploaded when the mutation actually runs.
        """
        self.files.append(file)

    def add_file_payloads(self, file_id: str, fill_payload):
        existing_files = [f.id for f in self.files if f]
        if file_id not in existing_files:
            raise ValueError(
                f"`{file_id}` was not added as a file. Please add "
                f"the file first before adding a fill payload."
            )
        self.file_payloads[file_id] = fill_payload

    def get_file_payloads(self):
        existing_files = [f.id for f in self.files if f]
        for key, _ in self.file_payloads.items():
            if key not in existing_files:
                raise ValueError(
                    f"`{key}` was not added as a file. Please add "
                    f"that file or remove its fill payload before "
                    f"attempting to create an Etch payload."
                )
        return self.file_payloads

    def get_files_for_payload(self) -> Tuple[List[AttachableEtchFile], List[Any]]:
        """Scan through and gather files for use in multipart requests.

        During the process, files will be replaced with a dummy `Path`
        instance as `Callable` instances and likely their returns of some
        sort of IO reader is not serializable and not easily dealt with in
        pydantic (the underlying model system for the types used in this
        library). Closer to the actual request, when the multipart payload is
        being constructed, the dummy instances will be replaced by their actual
        file instances.

        :return: Tuple containing a list of uploadable file types, and a list
            of actual file references
        :rtype: Tuple[List, List]
        """
        contains_uploads = any(  # pylint: disable=use-a-generator
            [isinstance(f, DocumentUpload) for f in self.files]
        )
        if not contains_uploads:
            return self.files, []

        files = []
        for f in self.files:
            if not isinstance(f, DocumentUpload):
                continue

            if getattr(f.Config, "contains_uploads", False):
                attr_name = getattr(f.Config, "contains_uploads")
                upload = getattr(f, attr_name)

                if not isinstance(upload, Base64Upload):
                    files.append(upload)
                    setattr(f, attr_name, Path())

        return self.files, files

    def create_payload(self) -> Tuple[CreateEtchPacketPayload, List]:
        """Create a payload based on data set on the class instance.

        Check `api_resources.payload.CreateEtchPacketPayload` for full payload
        requirements. Data requirements aren't explicitly enforced here, but
        at the payload class level.
        """
        # If there's an existing payload instance attribute, just return that.
        if self.payload:
            return self.payload, []

        if not self.name:
            raise TypeError("`name` and `signature_email_subject` cannot be None")

        payload_files, file_refs = self.get_files_for_payload()

        payload = CreateEtchPacketPayload(
            is_test=self.is_test,
            is_draft=self.is_draft,
            name=self.name,
            signers=self.signers,
            files=payload_files,
            data=CreateEtchFilePayload(payloads=self.get_file_payloads()),
            signature_email_subject=self.signature_email_subject,
            signature_email_body=self.signature_email_body,
            signature_page_options=self.signature_page_options or {},
            webhook_url=self.webhook_url,
            reply_to_email=self.reply_to_email,
            reply_to_name=self.reply_to_name,
            merge_pdfs=self.merge_pdfs,
            enable_emails=self.enable_emails,
            create_cast_templates_from_uploads=self.create_cast_templates_from_uploads,
            duplicate_casts=self.duplicate_casts,
        )

        return payload, file_refs



from python_anvil.api_resources.mutations.base import BaseQuery
from python_anvil.api_resources.payload import GenerateEtchSigningURLPayload


class GenerateEtchSigningURL(BaseQuery):
    """Query class to handle retrieving a signing URL."""

    mutation = """
        mutation ($signerEid: String!, $clientUserId: String!) {
            generateEtchSignURL (signerEid: $signerEid, clientUserId: $clientUserId)
        }
    """

    def __init__(self, signer_eid: str, client_user_id: str):
        self.signer_eid = signer_eid
        self.client_user_id = client_user_id

    def create_payload(self):
        return GenerateEtchSigningURLPayload(
            signer_eid=self.signer_eid, client_user_id=self.client_user_id
        )



import requests
from base64 import b64encode
from logging import getLogger
from ratelimit import limits, sleep_and_retry
from ratelimit.exception import RateLimitException
from requests.auth import HTTPBasicAuth

from python_anvil.exceptions import AnvilRequestException


logger = getLogger(__name__)

REQUESTS_LIMIT = {
    "dev": {
        "calls": 2,
        "seconds": 1,
    },
    "prod": {
        "calls": 40,
        "seconds": 1,
    },
}

RATELIMIT_ENV = "dev"


def _handle_request_error(e: Exception):
    raise e


class HTTPClient:
    def __init__(self, api_key=None, environment="dev"):
        self._session = requests.Session()
        self.api_key = api_key
        global RATELIMIT_ENV  # pylint: disable=global-statement
        RATELIMIT_ENV = environment

    def get_auth(self, encode=False) -> str:
        # TODO: Handle OAuth + API_KEY
        if not self.api_key:
            raise AttributeError("You must have an API key")

        # By default, the `requests` package will base64encode things with
        # the `HTTPBasicAuth` method, so no need to handle that here, but the
        # option is here if you _really_ want it.
        if encode:
            return b64encode(f"{self.api_key}:".encode()).decode()

        return self.api_key

    @sleep_and_retry
    @limits(
        calls=REQUESTS_LIMIT[RATELIMIT_ENV]["calls"],
        period=REQUESTS_LIMIT[RATELIMIT_ENV]["seconds"],
    )
    def do_request(
        self,
        method,
        url,
        headers=None,
        data=None,
        auth=None,
        params=None,
        retry=True,
        files=None,
        **kwargs,
    ) -> requests.Response:
        for _ in range(5):
            # Retry a max of 5 times in case of hitting any rate limit errors
            res = self._session.request(
                method,
                url,
                headers=headers,
                data=data,
                auth=auth,
                params=params,
                files=files,
                **kwargs,
            )

            if res.status_code == 429:
                time_to_wait = int(res.headers.get("Retry-After", 1))
                if retry:
                    logger.warning(
                        "Rate-limited: request not accepted. Retrying in "
                        "%i second%s.",
                        time_to_wait,
                        's' if time_to_wait > 1 else '',
                    )

                    # This exception will raise up to the `sleep_and_retry` decorator
                    # which will handle waiting for `time_to_wait` seconds.
                    raise RateLimitException("Retrying", period_remaining=time_to_wait)

                raise AnvilRequestException(
                    f"Rate limit exceeded. Retry after {time_to_wait} seconds."
                )

            break

        return res

    def request(
        self,
        method,
        url,
        headers=None,
        data=None,
        auth=None,
        params=None,
        retry=True,
        files=None,
        **kwargs,
    ):
        """Make an HTTP request.

        :param method: HTTP method to use
        :param url: URL to make the request on.
        :param headers:
        :param data:
        :param auth:
        :param params:
        :param files:
        :param retry: Whether to retry on any rate-limited requests
        :param kwargs:
        :return:
        """
        parse_json = kwargs.pop("parse_json", False)
        if self.api_key and not auth:
            auth = HTTPBasicAuth(self.get_auth(), "")

        try:
            res = self.do_request(
                method,
                url,
                headers=headers,
                data=data,
                auth=auth,
                params=params,
                retry=retry,
                files=files,
                **kwargs,
            )

            if parse_json and res.headers.get("Content-Type") == "application/json":
                content = res.json()
            else:
                # This actually reads the content and can potentially cause issues
                # depending on the content.
                # The structure of this method is very similar to Stripe's requests
                # HTTP client: https://github.com/stripe/stripe-python/blob/afa872c538bee0a1e14c8e131df52dd3c24ff05a/stripe/http_client.py#L304-L308
                content = res.content
            status_code = res.status_code
        except Exception as e:  # pylint: disable=broad-except
            _handle_request_error(e)

        return content, status_code, res.headers



# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned,singleton-comparison

from python_anvil.utils import build_batch_filenames, camel_to_snake, create_unique_id


def describe_build_batch_filenames():
    def test_with_normal_filename():
        gen = build_batch_filenames("somefile.txt")
        assert next(gen) == "somefile-0.txt"
        assert next(gen) == "somefile-1.txt"

    def test_with_start_index():
        gen = build_batch_filenames("somefile.txt", start_idx=100)
        assert next(gen) == "somefile-100.txt"
        assert next(gen) == "somefile-101.txt"

    def test_with_separator():
        gen = build_batch_filenames("somefile.txt", separator=":::")
        assert next(gen) == "somefile:::0.txt"

    def test_with_all():
        gen = build_batch_filenames("somefile.txt", start_idx=555, separator=":::")
        assert next(gen) == "somefile:::555.txt"


def describe_create_unique_id():
    def test_no_prefix():
        assert create_unique_id().startswith("field-")

    def test_prefix():
        prefix = "somePrefix+++--"
        assert create_unique_id(prefix).startswith(prefix)


def describe_camel_to_snake():
    def test_it():
        assert camel_to_snake('oneToTwo') == 'one_to_two'
        assert camel_to_snake('one') == 'one'
        assert camel_to_snake('TwoTwo') == 'two_two'



from pkg_resources import DistributionNotFound, get_distribution

from python_anvil import api, cli


try:
    __version__ = get_distribution('python_anvil').version
except DistributionNotFound:
    __version__ = '(local)'



from typing import Optional


class BaseQuery:
    """Base class for any GraphQL queries/mutations."""

    mutation: Optional[str] = None
    mutation_res_query: Optional[str] = None

    def get_mutation(self):
        if self.mutation and self.mutation_res_query:
            return self.mutation.format(query=self.mutation_res_query)
        return self.mutation

    def create_payload(self):
        if not self.mutation:
            raise ValueError(
                "`mutation` property must be set on the inheriting class level"
            )
        raise NotImplementedError()



from logging import getLogger
from typing import Any, AnyStr, Callable, Dict, List, Optional, Text, Tuple, Union

from .api_resources.mutations import (
    BaseQuery,
    CreateEtchPacket,
    ForgeSubmit,
    GenerateEtchSigningURL,
)
from .api_resources.payload import (
    CreateEtchPacketPayload,
    FillPDFPayload,
    ForgeSubmitPayload,
    GeneratePDFPayload,
)
from .api_resources.requests import GraphqlRequest, PlainRequest, RestRequest
from .http import HTTPClient
from .multipart_helpers import get_multipart_payload


logger = getLogger(__name__)


def _get_return(res: Dict, get_data: Callable[[Dict], Union[Dict, List]]):
    """Process response and get data from path if provided."""
    _res = res
    if "response" in res and "headers" in res:
        _res = res["response"]
        return {"response": get_data(_res), "headers": res["headers"]}
    return get_data(_res)


class Anvil:
    """Main Anvil API class.

    Handles all GraphQL and REST queries.

    Usage:
        >> anvil = Anvil(api_key="my_key")
        >> payload = {}
        >> pdf_data = anvil.fill_pdf("the_template_id", payload)
    """

    # Version number to use for latest versions (usually drafts)
    VERSION_LATEST = -1
    # Version number to use for the latest published version.
    # This is the default when a version is not provided.
    VERSION_LATEST_PUBLISHED = -2

    def __init__(self, api_key=None, environment='dev'):
        self.client = HTTPClient(api_key=api_key, environment=environment)

    def query(
        self,
        query: str,
        variables: Union[Optional[Text], Dict[Text, Any]] = None,
        **kwargs,
    ):
        gql = GraphqlRequest(client=self.client)
        return gql.post(query, variables=variables, **kwargs)

    def mutate(self, query: BaseQuery, variables: dict, **kwargs):
        gql = GraphqlRequest(client=self.client)
        return gql.post(query.get_mutation(), variables, **kwargs)

    def mutate_multipart(self, files: Dict, **kwargs):
        """
        Multipart version of `mutate`.

        This will send a mutation based on the multipart spec defined here:
        https://github.com/jaydenseric/graphql-multipart-request-spec

        Note that `variables` and `query` have been removed and replaced with
        `files`. The entire GraphQL payload should already be prepared as a
        `dict` beforehand.

        :param files:
        :param kwargs:
        :return:
        """
        gql = GraphqlRequest(client=self.client)
        return gql.post_multipart(files=files, **kwargs)

    def request_rest(self, options: Optional[dict] = None):
        api = RestRequest(self.client, options=options)
        return api

    def fill_pdf(
        self, template_id: str, payload: Union[dict, AnyStr, FillPDFPayload], **kwargs
    ):
        """Fill an existing template with provided payload data.

        Use the casts graphql query to get a list of available templates you
        can use for this request.

        :param template_id: eid of an existing template/cast
        :type template_id: str
        :param payload: payload in the form of a dict or JSON data
        :type payload: dict|str
        :param kwargs.version_number: specific template version number to use. If
            not provided, the latest _published_ version will be used.
        :type kwargs.version_number: int
        """
        try:
            if isinstance(payload, dict):
                data = FillPDFPayload(**payload)
            elif isinstance(payload, str):
                data = FillPDFPayload.parse_raw(
                    payload, content_type="application/json"
                )
            elif isinstance(payload, FillPDFPayload):
                data = payload
            else:
                raise ValueError("`payload` must be a valid JSON string or a dict")
        except KeyError as e:
            logger.exception(e)
            raise ValueError(
                "`payload` validation failed. Please make sure all required "
                "fields are set. "
            ) from e

        version_number = kwargs.pop("version_number", None)
        if version_number:
            kwargs["params"] = dict(versionNumber=version_number)

        api = RestRequest(client=self.client)
        return api.post(
            f"fill/{template_id}.pdf",
            data.dict(by_alias=True, exclude_none=True) if data else {},
            **kwargs,
        )

    def generate_pdf(self, payload: Union[AnyStr, Dict, GeneratePDFPayload], **kwargs):
        if not payload:
            raise ValueError("`payload` must be a valid JSON string or a dict")

        if isinstance(payload, dict):
            data = GeneratePDFPayload(**payload)
        elif isinstance(payload, str):
            data = GeneratePDFPayload.parse_raw(
                payload, content_type="application/json"
            )
        elif isinstance(payload, GeneratePDFPayload):
            data = payload
        else:
            raise ValueError("`payload` must be a valid JSON string or a dict")

        # Any data errors would come from here
        api = RestRequest(client=self.client)
        return api.post(
            "generate-pdf", data=data.dict(by_alias=True, exclude_none=True), **kwargs
        )

    def get_cast(
        self,
        eid: str,
        fields: Optional[List[str]] = None,
        version_number: Optional[int] = None,
        cast_args: Optional[List[Tuple[str, str]]] = None,
        **kwargs,
    ) -> Dict[str, Any]:

        if not fields:
            # Use default fields
            fields = ["eid", "title", "fieldInfo"]

        if not cast_args:
            cast_args = []

        cast_args.append(("eid", f'"{eid}"'))

        # If `version_number` isn't provided, the API will default to the
        # latest published version.
        if version_number:
            cast_args.append(("versionNumber", str(version_number)))

        arg_str = ""
        if len(cast_args):
            joined_args = [(":".join(arg)) for arg in cast_args]
            arg_str = f"({','.join(joined_args)})"

        res = self.query(
            f"""{{
              cast {arg_str} {{
                {" ".join(fields)}
              }}
            }}""",
            **kwargs,
        )

        def get_data(r) -> Dict[str, Any]:
            return r["data"]["cast"]

        return _get_return(res, get_data=get_data)

    def get_casts(self, fields=None, show_all=False, **kwargs) -> List[Dict[str, Any]]:
        if not fields:
            # Use default fields
            fields = ["eid", "title", "fieldInfo"]

        cast_args = "" if show_all else "(isTemplate: true)"

        res = self.query(
            f"""{{
              currentUser {{
                organizations {{
                  casts {cast_args} {{
                    {" ".join(fields)}
                  }}
                }}
              }}
            }}""",
            **kwargs,
        )

        def get_data(r):
            orgs = r["data"]["currentUser"]["organizations"]
            return [item for org in orgs for item in org["casts"]]

        return _get_return(res, get_data=get_data)

    def get_current_user(self, **kwargs):
        res = self.query(
            """{
              currentUser {
                name
                email
                eid
                role
                organizations {
                  eid
                  name
                  slug
                  casts {
                    eid
                    name
                  }
                }
              }
            }
            """,
            **kwargs,
        )

        return _get_return(res, get_data=lambda r: r["data"]["currentUser"])

    def get_welds(self, **kwargs) -> Union[List, Tuple[List, Dict]]:
        res = self.query(
            """{
              currentUser {
                organizations {
                  welds {
                    eid
                    slug
                    title
                    forges {
                      eid
                      name
                    }
                  }
                }
              }
            }""",
            **kwargs,
        )

        def get_data(r):
            orgs = r["data"]["currentUser"]["organizations"]
            return [item for org in orgs for item in org["welds"]]

        return _get_return(res, get_data=get_data)

    def get_weld(self, eid: Text, **kwargs):
        res = self.query(
            """
            query WeldQuery(
                #$organizationSlug: String!,
                #$slug: String!
                $eid: String!
            ) {
                weld(
                    #organizationSlug: $organizationSlug,
                    #slug: $slug
                    eid: $eid
                ) {
                    eid
                    slug
                    name
                    forges {
                        eid
                        name
                        slug
                    }
                }
            }""",
            variables=dict(eid=eid),
            **kwargs,
        )

        def get_data(r):
            return r["data"]["weld"]

        return _get_return(res, get_data=get_data)

    def create_etch_packet(
        self,
        payload: Optional[
            Union[
                dict,
                CreateEtchPacketPayload,
                CreateEtchPacket,
                AnyStr,
            ]
        ] = None,
        json=None,
        **kwargs,
    ):
        """Create etch packet via a graphql mutation."""
        # Create an etch packet payload instance excluding signers and files
        # (if any). We'll need to add those separately. below.
        if not any([payload, json]):
            raise TypeError('One of the arguments `payload` or `json` must exist')

        if json:
            payload = CreateEtchPacketPayload.parse_raw(
                json, content_type="application/json"
            )

        if isinstance(payload, dict):
            mutation = CreateEtchPacket.create_from_dict(payload)
        elif isinstance(payload, CreateEtchPacketPayload):
            mutation = CreateEtchPacket(payload=payload)
        elif isinstance(payload, CreateEtchPacket):
            mutation = payload
        else:
            raise ValueError(
                "`payload` must be a valid CreateEtchPacket instance or dict"
            )

        files = get_multipart_payload(mutation)

        res = self.mutate_multipart(files, **kwargs)
        return res

    def generate_etch_signing_url(self, signer_eid: str, client_user_id: str, **kwargs):
        """Generate a signing URL for a given user."""
        mutation = GenerateEtchSigningURL(
            signer_eid=signer_eid,
            client_user_id=client_user_id,
        )
        payload = mutation.create_payload()
        return self.mutate(mutation, variables=payload.dict(by_alias=True), **kwargs)

    def download_documents(self, document_group_eid: str, **kwargs):
        """Retrieve all completed documents in zip form."""
        api = PlainRequest(client=self.client)
        return api.get(f"document-group/{document_group_eid}.zip", **kwargs)

    def forge_submit(
        self,
        payload: Optional[Union[Dict[Text, Any], ForgeSubmitPayload]] = None,
        json=None,
        **kwargs,
    ):
        """Create a Webform (forge) submission via a graphql mutation."""
        if not any([json, payload]):
            raise TypeError('One of arguments `json` or `payload` are required')

        if json:
            payload = ForgeSubmitPayload.parse_raw(
                json, content_type="application/json"
            )

        if isinstance(payload, dict):
            mutation = ForgeSubmit.create_from_dict(payload)
        elif isinstance(payload, ForgeSubmitPayload):
            mutation = ForgeSubmit(payload=payload)
        else:
            raise ValueError(
                "`payload` must be a valid ForgeSubmitPayload instance or dict"
            )

        return self.mutate(
            mutation,
            variables=mutation.create_payload().dict(by_alias=True, exclude_none=True),
            **kwargs,
        )



ETCH_TEST_PAYLOAD = dict(
    name="Packet name",
    signature_email_subject="The subject",
    signers=[
        dict(
            name="Joe Doe",
            email="joe@example.com",
            fields=[dict(fileId="existingCast", fieldId="signMe")],
        )
    ],
    files=[
        dict(
            id="someFile",
            title="Sign This",
            file=dict(
                data="Some Base64 Thing",
                filename="someFile.pdf",
                mimetype="application/pdf",
            ),
            fields=[
                dict(
                    id="signField",
                    type="signature",
                    pageNum=0,
                    rect=dict(x=100, y=100, width=100, height=100),
                )
            ],
        )
    ],
)

EXPECTED_FILES = [
    {
        'id': 'someFile',
        'title': 'Sign This',
        'file': {
            'data': 'Some Base64 Thing',
            'filename': 'someFile.pdf',
            'mimetype': 'application/pdf',
        },
        'fields': [
            {
                'id': 'signField',
                'type': 'signature',
                'pageNum': 0,
                'rect': {'x': 100, 'y': 100, 'width': 100, 'height': 100},
            }
        ],
        'fontSize': 14,
        'textColor': '#000000',
    }
]

EXPECTED_ETCH_TEST_PAYLOAD = {
    'name': 'Packet name',
    'signatureEmailSubject': 'The subject',
    'signers': [
        {
            'name': 'Joe Doe',
            'email': 'joe@example.com',
            'fields': [{'fileId': 'existingCast', 'fieldId': 'signMe'}],
            'id': 'signer-mock-generated',
            'routingOrder': 1,
            'signerType': 'email',
        }
    ],
    'isDraft': False,
    'isTest': True,
    'data': {'payloads': {}},
    'signaturePageOptions': {},
    'files': EXPECTED_FILES,
}

EXPECTED_ETCH_TEST_PAYLOAD_JSON = {
    'name': 'Packet name',
    'signatureEmailSubject': 'The subject',
    'signers': [
        {
            'name': 'Joe Doe',
            'email': 'joe@example.com',
            'fields': [{'fileId': 'existingCast', 'fieldId': 'signMe'}],
            'id': '',
            'routingOrder': 1,
            'signerType': 'email',
        }
    ],
    'isDraft': False,
    'isTest': True,
    'data': None,
    'signaturePageOptions': None,
    'files': EXPECTED_FILES,
}



# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned,singleton-comparison
import pytest
from typing import Dict
from unittest import mock

from python_anvil.exceptions import AnvilRequestException
from python_anvil.http import HTTPClient


class HTTPResponse:
    status_code = 200
    content = ""
    headers: Dict = {}


def describe_http_client():
    @pytest.fixture
    def mock_response():
        class MockResponse:
            status_code = 429
            headers = {"Retry-After": 1}

        return MockResponse

    def test_client():
        client = HTTPClient()
        assert isinstance(client, HTTPClient)

    def describe_get_auth():
        def test_no_key():
            """Test that no key will raise an exception."""
            client = HTTPClient()
            with pytest.raises(AttributeError):
                client.get_auth()

        def test_key():
            key = "my_secret_ket!!!11!!"
            client = HTTPClient(api_key=key)
            assert client.get_auth() == key

        @mock.patch('python_anvil.http.b64encode')
        def test_encoded_key(mock_b64):
            key = "my_secret_ket!!!11!!"
            client = HTTPClient(api_key=key)
            client.get_auth(encode=True)
            mock_b64.assert_called_once()

    def describe_request():
        @mock.patch("python_anvil.http.HTTPBasicAuth")
        @mock.patch("python_anvil.http.HTTPClient.do_request")
        def test_default_args(do_request, basic_auth):
            basic_auth.return_value = "my_auth"
            response = HTTPResponse()
            do_request.return_value = response

            client = HTTPClient(api_key="my_key")
            res = client.request("GET", "http://localhost")

            assert res == (response.content, response.status_code, response.headers)
            do_request.assert_called_once_with(
                "GET",
                "http://localhost",
                headers=None,
                data=None,
                auth="my_auth",
                params=None,
                retry=True,
                files=None,
            )

    def describe_do_request():
        @mock.patch("python_anvil.http.requests.Session")
        def test_default_args(session):
            mock_session = mock.MagicMock()
            session.return_value = mock_session
            client = HTTPClient(api_key="my_key")
            client.do_request("GET", "http://localhost")

            # Should only be called once, never retried.
            mock_session.request.assert_called_once_with(
                "GET",
                "http://localhost",
                headers=None,
                data=None,
                auth=None,
                params=None,
                files=None,
            )

        @mock.patch("python_anvil.http.RateLimitException")
        @mock.patch("python_anvil.http.requests.Session")
        def test_default_args_with_retry(session, ratelimit_exc, mock_response):
            class MockException(Exception):
                pass

            mock_session = mock.MagicMock()
            mock_session.request.return_value = mock_response()
            session.return_value = mock_session
            ratelimit_exc.return_value = MockException()

            client = HTTPClient(api_key="my_key")
            with pytest.raises(MockException):
                client.do_request("GET", "http://localhost", retry=True)

            assert ratelimit_exc.call_count == 1

            # Should only be called once, would retry but RateLimitException
            # is mocked here.
            mock_session.request.assert_called_once_with(
                "GET",
                "http://localhost",
                headers=None,
                data=None,
                auth=None,
                params=None,
                files=None,
            )

        @mock.patch("python_anvil.http.RateLimitException")
        @mock.patch("python_anvil.http.requests.Session")
        def test_default_args_without_retry(session, ratelimit_exc, mock_response):
            class MockException(Exception):
                pass

            mock_session = mock.MagicMock()
            mock_session.request.return_value = mock_response()
            session.return_value = mock_session
            ratelimit_exc.return_value = MockException()

            client = HTTPClient(api_key="my_key")
            with pytest.raises(AnvilRequestException):
                client.do_request("GET", "http://localhost", retry=False)

            assert ratelimit_exc.call_count == 0

            # Should only be called once, never retried.
            mock_session.request.assert_called_once_with(
                "GET",
                "http://localhost",
                headers=None,
                data=None,
                auth=None,
                params=None,
                files=None,
            )



import re
import uuid
from logging import getLogger
from os import path


logger = getLogger(__name__)


def build_batch_filenames(filename: str, start_idx=0, separator: str = "-"):
    """
    Create a generator for filenames in sequential order.

    Example:
        build_batch_filenames('somefile.pdf') will yield filenames:
        * somefile-1.pdf
        * somefile-2.pdf
        * somefile-3.pdf
    :param filename: Full filename, including extension
    :param start_idx: Starting index number
    :param separator:
    :return:
    """
    idx = start_idx or 0
    sep = separator or "-"
    file_part, ext = path.splitext(filename)

    while True:
        yield f"{file_part}{sep}{idx}{ext}"
        idx += 1


def create_unique_id(prefix: str = "field") -> str:
    """Create a prefixed unique id."""
    return f"{prefix}-{uuid.uuid4().hex}"


def remove_empty_items(dict_obj: dict):
    """Remove null values from a dict."""
    return {k: v for k, v in dict_obj.items() if v is not None}


def camel_to_snake(name: str) -> str:
    return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()



import copy
import json
import mimetypes
from io import BufferedIOBase
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from python_anvil.api_resources.mutations import BaseQuery


FileLikeObject = Union[Path, BufferedIOBase]


def get_extractable_files_from_payload(
    variables: Any,
    is_match: Callable,
    cur_path: Optional[str] = None,
    cur_files: Optional[List] = None,
) -> Tuple[List[Tuple[str, FileLikeObject]], bool]:
    """Process variables bound for a Graphql multipart request.

    IMPORTANT: This will __overwrite__ the file-like objects to `None` due to
    the expected request format in the spec.

    Spec: https://github.com/jaydenseric/graphql-multipart-request-spec

    :param variables:
    :param is_match:
    :param cur_path:
    :param cur_files:
    :return: List containing a tuple of (file path string, file path/buffer object)
    :rtype: Tuple[List[Tuple[str, Union[Path, BufferedIOBase]], bool, List]
    """

    if cur_path is None:
        cur_path = "variables"
    if cur_files is None:
        cur_files = []
    if not variables:
        return [], False

    if is_match(variables):
        cur_files.append((cur_path, copy.deepcopy(variables)))
        return cur_files, True

    if isinstance(variables, dict):
        to_remove = []
        for key in variables:
            _, remove = get_extractable_files_from_payload(
                variables[key],
                is_match=is_match,
                cur_path=f"{cur_path}.{key}",
                cur_files=cur_files,
            )
            if remove:
                to_remove.append(key)

        for key in to_remove:
            variables[key] = None

    elif isinstance(variables, list):
        to_remove = []
        for idx, item in enumerate(variables):
            _, remove = get_extractable_files_from_payload(
                item,
                is_match=is_match,
                cur_path=f"{cur_path}.{idx}",
                cur_files=cur_files,
            )
            if remove:
                to_remove.append(idx)

        for idx in to_remove:
            variables[idx] = None

    return cur_files, False


def get_multipart_payload(mutation: BaseQuery):  # pylint: disable=too-many-locals
    def is_match(item):
        return isinstance(item, (BufferedIOBase, Path))

    payload, to_upload = mutation.create_payload()
    variables = payload.dict(by_alias=True, exclude_none=True)
    multipart_map, _ = get_extractable_files_from_payload(variables, is_match=is_match)

    operations = json.dumps(
        {
            "query": mutation.get_mutation(),
            "variables": variables,
        }
    )

    filemap = {}
    for idx, filepath in enumerate(multipart_map):
        filemap[idx] = [filepath[0]]

    files = {
        "operations": (None, operations),
        "map": (None, json.dumps(filemap)),
    }  # type: Dict[str, Union[Tuple[None, str], Tuple[Any, Any, Optional[str]]]]

    for idx, file_or_path in enumerate(to_upload):
        # `key` here is most likely going to be a list index (int), but
        # `requests` will expect an actual string when it constructs the
        # multipart request. We make sure this is a string here.
        actual_key = str(idx)

        # This is already a file-like object, pass it directly to `requests`.
        if isinstance(file_or_path, Callable):  # type: ignore
            # If you have a `ValueError` here, make sure your callable returns
            # a tuple of the file and filename:
            # Example: (open(file, "rb"), "file.pdf"))
            file_part, name_part = file_or_path()
        elif isinstance(file_or_path, Path):
            file_part = open(file_or_path, "rb")  # pylint: disable=consider-using-with
            name_part = file_or_path.parts[-1]
        else:
            raise AssertionError("File path or file-like object not given")

        mimetype, _ = mimetypes.guess_type(file_part.name)
        files[actual_key] = (name_part, file_part, mimetype)

    return files



from typing import List, Text, Type

from python_anvil.api_resources.base import BaseModel


def get_payload_attrs(payload_model: Type[BaseModel]) -> List[Text]:
    return list(payload_model.__fields__.keys())



from typing import Any, Dict, Optional, Text, Union

from python_anvil.api_resources.mutations.base import BaseQuery
from python_anvil.api_resources.mutations.helpers import get_payload_attrs
from python_anvil.api_resources.payload import ForgeSubmitPayload


DEFAULT_RESPONSE_QUERY = """
{
  id
  eid
  status
  resolvedPayload
  currentStep
  completedAt
  createdAt
  updatedAt
  signer {
    name
    email
    status
    routingOrder
  }
  weldData {
    id
    eid
    status
    isTest
    isComplete
    agents
  }
}
"""

# NOTE: Since the below will be used as a formatted string (this also applies
#   to f-strings) any literal curly braces need to be doubled, else they'll be
#   interpreted as string replacement tokens.
FORGE_SUBMIT = """
mutation ForgeSubmit(
    $forgeEid: String!,
    $weldDataEid: String,
    $submissionEid: String,
    $payload: JSON!,
    $currentStep: Int,
    $complete: Boolean,
    $isTest: Boolean,
    $timezone: String,
    $groupArrayId: String,
    $groupArrayIndex: Int,
    $errorType: String,
    $webhookURL: String,
) {{
    forgeSubmit (
        forgeEid: $forgeEid,
        weldDataEid: $weldDataEid,
        submissionEid: $submissionEid,
        payload: $payload,
        currentStep: $currentStep,
        complete: $complete,
        isTest: $isTest,
        timezone: $timezone,
        groupArrayId: $groupArrayId,
        groupArrayIndex: $groupArrayIndex,
        errorType: $errorType,
        webhookURL: $webhookURL
    ) {query}
}}
"""


class ForgeSubmit(BaseQuery):
    mutation = FORGE_SUBMIT
    mutation_res_query = DEFAULT_RESPONSE_QUERY

    def __init__(
        self,
        payload: Union[Dict[Text, Any], ForgeSubmitPayload],
        forge_eid: Optional[Text] = None,
        weld_data_eid: Optional[Text] = None,
        submission_eid: Optional[Text] = None,
        is_test: Optional[bool] = None,
        **kwargs,
    ):
        """
        Create a forgeSubmit query.

        :param forge_eid:
        :param payload:
        :param weld_data_eid:
        :param submission_eid:
        :param is_test:
        :param kwargs: kwargs may contain other fields defined in
              `ForgeSubmitPayload` if not explicitly in the `__init__` args.
        """
        if not forge_eid and not isinstance(payload, ForgeSubmitPayload):
            raise ValueError(
                "`forge_eid` is required if `payload` is not a "
                "`ForgeSubmitPayload` instance"
            )

        self.payload = payload
        self.forge_eid = forge_eid
        self.weld_data_eid = weld_data_eid
        self.submission_eid = submission_eid
        self.is_test = is_test

        # Get other attrs from the model and set on the instance
        model_attrs = get_payload_attrs(ForgeSubmitPayload)
        for attr in model_attrs:
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])

    @classmethod
    def create_from_dict(cls, payload: Dict[Text, Any]):
        # Parse the data through the model class to validate and pass it back
        # as variables in this class.
        return cls(**payload)

    def create_payload(self):
        # If provided a payload and no forge_eid, we'll assume that it's the
        # full thing. Return that instead.
        if not self.forge_eid and self.payload:
            return self.payload

        model_attrs = get_payload_attrs(ForgeSubmitPayload)

        for_payload = {}
        for attr in model_attrs:
            obj = getattr(self, attr, None)
            if obj is not None:
                for_payload[attr] = obj

        return ForgeSubmitPayload(**for_payload)



# pylint: disable=unused-variable,unused-argument,too-many-statements
import json
import pytest
from pydantic import ValidationError
from typing import Any, MutableMapping
from unittest import mock

from python_anvil.api import Anvil, CreateEtchPacket
from python_anvil.api_resources.payload import (
    CreateEtchPacketPayload,
    ForgeSubmitPayload,
)

from ..api_resources.payload import FillPDFPayload
from . import payloads


DEV_KEY = "MY-SECRET-KEY"


def describe_api():
    @pytest.fixture
    def anvil():
        return Anvil(api_key=DEV_KEY)

    def describe_init():
        @mock.patch('python_anvil.api.HTTPClient')
        def test_init_key_default(mock_client):
            Anvil(api_key="what")
            mock_client.assert_called_once_with(api_key="what", environment="dev")

        @mock.patch('python_anvil.api.HTTPClient')
        def test_init_key_prod(mock_client):
            Anvil(api_key="what", environment="prod")
            mock_client.assert_called_once_with(api_key="what", environment="prod")

    def describe_query():
        def test_query():
            # TODO: ...
            pass

    def describe_fill_pdf():
        @mock.patch('python_anvil.api.RestRequest.post')
        def test_dict_payload(m_request_post, anvil):
            payload = {"data": {"this_data": "yes"}}
            anvil.fill_pdf("some_template", payload=payload)
            m_request_post.assert_called_once_with("fill/some_template.pdf", payload)

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_json_payload(m_request_post, anvil):
            payload = """{ "data": {"jsonData": "is here"} }"""
            anvil.fill_pdf("some_template", payload=payload)
            m_request_post.assert_called_once_with(
                "fill/some_template.pdf",
                {'data': {'jsonData': 'is here'}},
            )

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_payload_obj(m_request_post, anvil):
            payload = FillPDFPayload(data={"jsonData": "this was a payload instance"})
            anvil.fill_pdf("some_template", payload=payload)
            m_request_post.assert_called_once_with(
                "fill/some_template.pdf",
                {'data': {'jsonData': 'this was a payload instance'}},
            )

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_empty_payload(m_request_post, anvil):
            with pytest.raises(ValueError):
                anvil.fill_pdf("some_template", payload={})
            assert m_request_post.call_count == 0

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_with_kwargs(m_request_post, anvil):
            payload = """{ "data": {"jsonData": "is here"} }"""
            anvil.fill_pdf("some_template", payload=payload, include_headers=True)
            m_request_post.assert_called_once_with(
                "fill/some_template.pdf",
                {"data": {"jsonData": "is here"}},
                include_headers=True,
            )

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_with_version(m_request_post, anvil):
            payload = {"data": {"one": "One string"}}
            anvil.fill_pdf(
                "some_template",
                payload=payload,
                include_headers=True,
                version_number=Anvil.VERSION_LATEST,
            )
            m_request_post.assert_called_once_with(
                "fill/some_template.pdf",
                {"data": {"one": "One string"}},
                include_headers=True,
                params={"versionNumber": Anvil.VERSION_LATEST},
            )

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_with_params(m_request_post, anvil):
            payload = {"data": {"one": "One string"}}
            params = {"arbitrary": "Param"}
            anvil.fill_pdf(
                "some_template", payload=payload, include_headers=True, params=params
            )
            m_request_post.assert_called_once_with(
                "fill/some_template.pdf",
                {"data": {"one": "One string"}},
                include_headers=True,
                params=params,
            )

    def describe_generate_pdf():
        @mock.patch('python_anvil.api.RestRequest.post')
        def test_dict_payload(m_request_post, anvil):
            anvil.generate_pdf({"data": [{"d1": "data"}]})
            m_request_post.assert_called_once_with(
                # Defaults to 'markdown'
                "generate-pdf",
                data={'data': [{'d1': 'data'}], 'type': 'markdown'},
            )

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_json_payload(m_request_post, anvil):
            payload = """{ "data": [{ "d1": "data" }] }"""
            anvil.generate_pdf(payload)
            m_request_post.assert_called_once_with(
                "generate-pdf", data={"data": [{"d1": "data"}], "type": "markdown"}
            )

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_payload_html_type(m_request_post, anvil):
            anvil.generate_pdf({"data": {"html": "<h1>Hello</h1>"}, "type": "html"})
            m_request_post.assert_called_once_with(
                "generate-pdf",
                data={"data": {"html": "<h1>Hello</h1>"}, "type": "html"},
            )

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_invalid_payload_html_payload(m_request_post, anvil):
            with pytest.raises(ValueError):
                anvil.generate_pdf({"data": {"no_html_here": "Nope"}, "type": "html"})
            assert m_request_post.call_count == 0

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_payload_invalid_type(m_request_post, anvil):
            with pytest.raises(ValueError):
                anvil.generate_pdf(
                    {"data": [{"d1": "data"}], "type": "something_invalid"}
                )
            assert m_request_post.call_count == 0

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_invalid_data_for_html(m_request_post, anvil):
            with pytest.raises(ValueError):
                anvil.generate_pdf(
                    {
                        # This should be a plain dict, not a list
                        "data": [{"d1": "data"}],
                        "type": "html",
                    }
                )
            assert m_request_post.call_count == 0

        @mock.patch('python_anvil.api.RestRequest.post')
        def test_invalid_data_for_markdown(m_request_post, anvil):
            with pytest.raises(ValueError):
                anvil.generate_pdf(
                    {
                        # This should be a plain dict, not a list
                        "data": {"d1": "data"},
                        "type": "markdown",
                    }
                )
            assert m_request_post.call_count == 0

    def describe_current_user_query():
        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_get_current_user(m_request_post, anvil):
            anvil.get_current_user()
            assert m_request_post.call_count == 1

    def describe_get_welds():
        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_get_weld(m_request_post, anvil):
            anvil.get_welds()
            assert m_request_post.call_count == 1

    def describe_generate_etch_signing_url():
        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_get_url(m_request_post, anvil):
            anvil.generate_etch_signing_url(
                signer_eid='someId', client_user_id='anotherId'
            )
            assert m_request_post.call_count == 1

    def describe_download_documents():
        @mock.patch('python_anvil.api.PlainRequest.get')
        def test_get_docs(m_request_post, anvil):
            anvil.download_documents('someEid')
            assert m_request_post.call_count == 1

    def describe_get_cast():
        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_get_cast(m_request_post, anvil):
            anvil.get_cast('castEid')
            assert m_request_post.call_count == 1

    def describe_get_casts():
        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_get_casts(m_request_post, anvil):
            anvil.get_casts()
            assert m_request_post.call_count == 1

    def describe_create_etch_packet():
        expected_data = {
            'name': 'Packet name',
            'signatureEmailSubject': 'The subject',
            'signers': [],
            'files': [],
            'isDraft': False,
            'isTest': True,
            'data': {'payloads': {}},
            'signaturePageOptions': {},
        }

        @mock.patch('python_anvil.api.GraphqlRequest.post_multipart')
        def test_create_etch_packet_empty_payload(m_request_post, anvil):
            with pytest.raises(TypeError):
                anvil.create_etch_packet(payload={})
                assert m_request_post.call_count == 0

        @mock.patch('python_anvil.api.GraphqlRequest.post_multipart')
        def test_create_etch_packet_invalid_payload(m_request_post, anvil):
            with pytest.raises(TypeError):
                anvil.create_etch_packet({})
                assert m_request_post.call_count == 0

        @mock.patch('python_anvil.api.GraphqlRequest.post_multipart')
        def test_create_etch_packet_valid_payload_type(m_request_post, anvil):
            payload = CreateEtchPacket(
                name="Packet name",
                signature_email_subject="The subject",
            )
            anvil.create_etch_packet(payload=payload)
            assert m_request_post.call_count == 1

            files = m_request_post.call_args[1]["files"]
            files_payload = json.loads(files["operations"][1])
            variables = files_payload["variables"]
            assert expected_data == variables

        @mock.patch('python_anvil.api.GraphqlRequest.post_multipart')
        def test_create_etch_packet_passes_options(m_request_post, anvil):
            payload = CreateEtchPacket(
                is_test=False,
                is_draft=True,
                name="Packet name",
                signature_email_subject="The subject",
            )
            new_expected = dict(**expected_data)
            new_expected["isTest"] = False
            new_expected["isDraft"] = True
            anvil.create_etch_packet(payload)

            assert m_request_post.call_count == 1
            files = m_request_post.call_args[1]["files"]
            files_payload = json.loads(files["operations"][1])
            variables = files_payload["variables"]
            assert new_expected == variables

        @mock.patch('python_anvil.api.GraphqlRequest.post_multipart')
        def test_create_etch_packet_valid_dict_type(m_request_post, anvil):
            anvil.create_etch_packet(
                dict(name="Packet name", signature_email_subject="The subject")
            )
            assert m_request_post.call_count == 1
            files = m_request_post.call_args[1]["files"]
            files_payload = json.loads(files["operations"][1])
            variables = files_payload["variables"]
            assert expected_data == variables

        @mock.patch(
            'python_anvil.api_resources.mutations.create_etch_packet.create_unique_id'
        )
        @mock.patch('python_anvil.api.GraphqlRequest.post_multipart')
        def test_create_etch_packet_dict_with_signer(
            m_request_post, m_create_unique, anvil
        ):
            m_create_unique.return_value = "signer-mock-generated"
            anvil.create_etch_packet(payload=payloads.ETCH_TEST_PAYLOAD)
            assert m_request_post.call_count == 1
            files = m_request_post.call_args[1]["files"]
            files_payload = json.loads(files["operations"][1])
            variables = files_payload["variables"]
            assert payloads.EXPECTED_ETCH_TEST_PAYLOAD == variables

        @mock.patch(
            'python_anvil.api_resources.mutations.create_etch_packet.create_unique_id'
        )
        @mock.patch('python_anvil.api.GraphqlRequest.post_multipart')
        def test_create_etch_packet_json(m_request_post, m_create_unique, anvil):
            m_create_unique.return_value = "signer-mock-generated"
            # We are currently removing `None`s from the payload, so do that here too.
            payload = {
                k: v
                for k, v in payloads.EXPECTED_ETCH_TEST_PAYLOAD_JSON.items()
                if v is not None
            }
            anvil.create_etch_packet(json=json.dumps(payload))
            assert m_request_post.call_count == 1

            files = m_request_post.call_args[1]["files"]
            files_payload = json.loads(files["operations"][1])
            variables = files_payload["variables"]
            assert payload == variables

        @mock.patch(
            'python_anvil.api_resources.mutations.create_etch_packet.create_unique_id'
        )
        @mock.patch('python_anvil.api.GraphqlRequest.post_multipart')
        def test_adding_unsupported_fields(m_request_post, m_create_unique, anvil):
            m_create_unique.return_value = "signer-mock-generated"
            # We are currently removing `None`s from the payload, so do that here too.
            payload = {
                k: v
                for k, v in payloads.EXPECTED_ETCH_TEST_PAYLOAD_JSON.items()
                if v is not None
            }
            cep = CreateEtchPacketPayload.parse_obj(payload)
            cep.newFeature = True  # type: ignore[attr-defined]
            anvil.create_etch_packet(payload=cep)
            assert m_request_post.call_count == 1

            files = m_request_post.call_args[1]["files"]
            files_payload = json.loads(files["operations"][1])
            variables = files_payload["variables"]
            assert cep.dict(by_alias=True, exclude_none=True) == variables
            assert "newFeature" in variables

    def describe_forge_submit():
        expected_data = {
            "forgeEid": "forge1234",
            "payload": {"field1": "Some data"},
            "isTest": True,
        }

        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_invalid_payload(m_request_post, anvil):
            with pytest.raises(TypeError):
                anvil.forge_submit({})
                assert m_request_post.call_count == 0

        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_minimum_valid_data_forge_eid(m_request_post, anvil):
            payload = ForgeSubmitPayload(
                forge_eid="forge1234", payload=dict(field1="Some data")
            )
            anvil.forge_submit(payload=payload)
            assert m_request_post.call_count == 1
            assert expected_data == m_request_post.call_args[0][1]

        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_invalid_wd_submission(m_request_post, anvil):
            with pytest.raises(ValidationError):
                payload = dict(
                    forge_eid="forge1234",
                    # weld_data_eid and submission_eid must be provided
                    # if one exists.
                    weld_data_eid="wd1234",
                    payload=dict(field1="Updated data"),
                )
                anvil.forge_submit(payload=payload)
                assert m_request_post.call_count == 0

            with pytest.raises(ValidationError):
                payload = dict(
                    forge_eid="forge1234",
                    # weld_data_eid and submission_eid must be provided
                    # if one exists.
                    submission_eid="sub1234",
                    payload=dict(field1="Updated data"),
                )
                anvil.forge_submit(payload=payload)
                assert m_request_post.call_count == 0

        @mock.patch('python_anvil.api.GraphqlRequest.post')
        def test_minimum_valid_data_submission(m_request_post, anvil):
            payload = ForgeSubmitPayload(
                forge_eid="forge1234",
                submission_eid="sub1234",
                weld_data_eid="wd1234",
                payload=dict(field1="Updated data"),
            )

            # We copy `expected_data` here as it can cause a race condition
            # in other tests that use it.
            _expected_data: MutableMapping[str, Any] = {
                **expected_data,
                "submissionEid": "sub1234",
                "weldDataEid": "wd1234",
                "payload": {"field1": "Updated data"},
            }

            anvil.forge_submit(payload=payload)
            assert m_request_post.call_count == 1
            assert _expected_data in m_request_post.call_args[0]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['python_anvil',
 'python_anvil.api_resources',
 'python_anvil.api_resources.mutations',
 'python_anvil.tests']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0,<9.0',
 'pydantic>=1.8.2,<2.0.0',
 'ratelimit>=2.2.1,<3.0.0',
 'requests>=2.25.0,<3.0.0',
 'tabulate>=0.9.0,<0.10.0']

entry_points = \
{'console_scripts': ['anvil = python_anvil.cli:cli']}

setup_kwargs = {
    'name': 'python-anvil',
    'version': '2.0.0',
    'description': 'Anvil API',
    'long_description': '![Horizontal Lockupblack](https://user-images.githubusercontent.com/293079/169453889-ae211c6c-7634-4ccd-8ca9-8970c2621b6f.png#gh-light-mode-only)\n![Horizontal Lockup copywhite](https://user-images.githubusercontent.com/293079/169453892-895f637b-4633-4a14-b997-960c9e17579b.png#gh-dark-mode-only)\n\n# Anvil API Library\n\n[![PyPI Version](https://img.shields.io/pypi/v/python-anvil.svg)](https://pypi.org/project/python-anvil)\n[![PyPI License](https://img.shields.io/pypi/l/python-anvil.svg)](https://pypi.org/project/python-anvil)\n\nThis is a library that provides an interface to access the [Anvil API](https://www.useanvil.com/developers) from applications\nwritten in the Python programming language.\n\n[Anvil](https://www.useanvil.com/developers/) provides easy APIs for all things paperwork.\n\n1. [PDF filling API](https://www.useanvil.com/products/pdf-filling-api/) - fill out a PDF template with a web request and structured JSON data.\n2. [PDF generation API](https://www.useanvil.com/products/pdf-generation-api/) - send markdown or HTML and Anvil will render it to a PDF.\n3. [Etch e-sign with API](https://www.useanvil.com/products/etch/) - customizable, embeddable, e-signature platform with an API to control the signing process end-to-end.\n4. [Anvil Workflows (w/ API)](https://www.useanvil.com/products/workflows/) - Webforms + PDF + e-sign with a powerful no-code builder. Easily collect structured data, generate PDFs, and request signatures.\n\nLearn more on our [Anvil developer page](https://www.useanvil.com/developers/). See the [API guide](https://www.useanvil.com/docs) and the [GraphQL reference](https://www.useanvil.com/docs/api/graphql/reference/) for full documentation.\n\n### Documentation\n\nGeneral API documentation: [Anvil API docs](https://www.useanvil.com/docs)\n\n# Setup\n\n## Requirements\n\n* Python 3.6.2+\n\n## Installation\n\nInstall it directly into an activated virtual environment:\n\n```shell\n$ pip install python-anvil\n```\n\nor add it to your [Poetry](https://python-poetry.org/) project:\n\n```shell\n$ poetry add python-anvil\n```\n',
    'author': 'Anvil Foundry Inc.',
    'author_email': 'developers@useanvil.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.useanvil.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<3.12',
}


setup(**setup_kwargs)



# pylint: disable=no-self-argument
import re

# Disabling pylint no-name-in-module because this is the documented way to
# import `BaseModel` and it's not broken, so let's keep it.
from pydantic import BaseModel as _BaseModel, Extra  # pylint: disable=no-name-in-module


under_pat = re.compile(r"_([a-z])")


def underscore_to_camel(name):
    ret = under_pat.sub(lambda x: x.group(1).upper(), name)
    return ret


class BaseModel(_BaseModel):
    class Config:
        """Config override for all models.

        This override is mainly so everything can go from snake to camel-case.
        """

        alias_generator = underscore_to_camel
        allow_population_by_field_name = True

        # Allow extra fields even if it is not defined. This will allow models
        # to be more flexible if features are added in the Anvil API, but
        # explicit support hasn't been added yet to this library.
        extra = Extra.allow



# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned,singleton-comparison

from python_anvil.multipart_helpers import get_extractable_files_from_payload


def describe_get_extractable_files_from_payload():
    def test_it_works():
        def is_match(x):
            return isinstance(x, int)

        variables = {
            "one": [{"f1": "data", "f2": 1}],
            "two": {"s2-1": "more data", "s2-2": "more more data", "s2-3": 2},
            "three": "nothing here",
            "four": 4,
        }

        paths, _ = get_extractable_files_from_payload(
            variables, is_match, cur_path="variables"
        )
        assert len(paths) == 3
        assert ("variables.one.0.f2", 1) in paths
        assert ("variables.two.s2-3", 2) in paths
        assert ("variables.four", 4) in paths

    def test_handles_empty_things():
        payload_expected_pairs = [
            ({}, 0),
            ([], 0),
            ('', 0),
            (' ', 1),
        ]

        for var_input, exp_len in payload_expected_pairs:
            paths, _ = get_extractable_files_from_payload(
                var_input, lambda x: x, cur_path="variables"
            )
            assert len(paths) == exp_len
